//---------------------------------------------------------------------------

#include <vcl.h>
#pragma hdrstop

#include <ipl.h>
#include <cxcore.h>
#include <cv.h>
#include <highgui.h>

#include "Unit1.h"
#include <math.h>
//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma link "CSPIN"
#pragma resource "*.dfm"
TForm1 *Form1;
//---------------------------------------------------------------------------
__fastcall TForm1::TForm1(TComponent* Owner)
        : TForm(Owner)
{
}

//---------------------------------------------------------------------------
//  Definiciones y funciones auxiliares
//---------------------------------------------------------------------------

#define BLOQUE 31

bool izquierda;            // Direccin de giro, izquierda o derecha

#define VUELTA(A) {if (A->origin) {cvFlip(A); A->origin= 0;}}

void pintaFlecha (IplImage *res, int x1, int y1, int x0, int y0)
// Pinta una flecha en la imagen res, desde el punto (x1,y1) al (x0,y0)
{
  double a= 0.6, b= 0.3;
  int x2, y2;
  cvLine(res, cvPoint(x0,y0), cvPoint(x1,y1), CV_RGB(255,0,0), 1, CV_AA);
  x2= a*x1+(1.0-a)*x0+b*(y1-y0);
  y2= a*y1+(1.0-a)*y0+b*(x0-x1);
  cvLine(res, cvPoint(x2,y2), cvPoint(x1,y1), CV_RGB(255,0,0), 1, CV_AA);
  x2= a*x1+(1.0-a)*x0+b*(y0-y1);
  y2= a*y1+(1.0-a)*y0+b*(x1-x0);
  cvLine(res, cvPoint(x2,y2), cvPoint(x1,y1), CV_RGB(255,0,0), 1, CV_AA);
}

//---------------------------------------------------------------------------
//  Clculo de la mediana de los flujos calculados por bloques
//---------------------------------------------------------------------------

#define MAX_MEDIANA 200

float medx[MAX_MEDIANA];   // Valores de flujo en X
float medy[MAX_MEDIANA];   // Valores de flujo en Y
int nummed= 0;             // Nmero de valores introducidos

//-------------------------

void inic_mediana (void)
// Inicializa el clculo de la mediana
{
  nummed= 0;
}

//-------------------------

void anade_med (float x, float y)
// Aade un bloque a la mediana
{
  int i= 0;
  int j;
  if (nummed>=MAX_MEDIANA)
    return;
  // 1. Insercin ordenada de x en medx
  while (i<nummed && medx[i]<=x)
    i++;
  if (i==nummed)
    medx[nummed]= x;
  else {
    for (j= nummed; j>i; j--)
      medx[j]= medx[j-1];
    medx[i]= x;
  }
  // 2. Insercin ordenada de y en medy
  i= 0;
  while (i<nummed && medy[i]<=y)
    i++;
  if (i==nummed)
    medy[nummed]= y;
  else {
    for (j= nummed; j>i; j--)
      medy[j]= medy[j-1];
    medy[i]= y;
  }
  nummed++;
}

//-------------------------

float medianaX (void)
// Devuelve la mediana en X
{
  return medx[nummed/2];
}

//-------------------------

float medianaY (void)
// Devuelve la mediana en Y
{
  return medy[nummed/2];
}

//---------------------------------------------------------------------------
//  Clculo del flujo ptico por bloques
//---------------------------------------------------------------------------

void flujoBM (IplImage *src, IplImage *ant, double &dx, double &dy)
// Calcula el flujo ptico (aplicando un algoritmo de matching de bloques)
// entre las imgenes src y ant, devolviendo la mediana de los flujos
// resultantes en (dx, dy)
{
  int x, y;
  IplImage *srcg, *antg;      // Imgenes de entrada en escala de grises
  IplImage *velx, *vely;      // Mapas de flujos en X e Y

  // 1. Crear imgenes auxiliares
  srcg= cvCreateImage(cvGetSize(src), src->depth, 1);
  cvCvtColor(src, srcg, CV_BGR2GRAY);
  antg= cvCreateImage(cvGetSize(ant), ant->depth, 1);
  cvCvtColor(ant, antg, CV_BGR2GRAY);
  velx= cvCreateImage(cvSize((srcg->width+BLOQUE-1)/BLOQUE,
                             (srcg->height+BLOQUE-1)/BLOQUE),
                      IPL_DEPTH_32F, 1);
  vely= cvCloneImage(velx);

  // 2. Calcular el flujo ptico
  cvCalcOpticalFlowBM(srcg, antg, cvSize(BLOQUE,BLOQUE), cvSize(1,1),
                      cvSize(BLOQUE,BLOQUE/2), 0, velx, vely);

  // 3. Dibujar el flujo ptico (slo para visualizacin)
  IplImage *res= cvCloneImage(src);
  for (y= 1; y<velx->height-1; y+= 1)
    for (x= 1; x<vely->width-1; x+= 1) {
      dx= cvGetReal2D(velx, y, x);
      dy= cvGetReal2D(vely, y, x);
      pintaFlecha(res, 5+x*BLOQUE-dx, 5+y*BLOQUE-dy, 5+x*BLOQUE,5+y*BLOQUE);
    }
  cvNamedWindow("Flujo ptico", 1);
  cvShowImage("Flujo ptico", res);
  cvReleaseImage(&res);

  // 4. Calcular la mediana de los flujos resultantes
  inic_mediana();
  for (y= 1; y<velx->height-1; y+= 1)
    for (x= 1; x<vely->width-1; x+= 1) {
      dx= cvGetReal2D(velx, y, x);
      dy= cvGetReal2D(vely, y, x);
      anade_med(dx, dy);
    }
  dx= medianaX();
  dy= medianaY();

  // 5. Liberar las imgenes auxiliares
  cvReleaseImage(&srcg);
  cvReleaseImage(&antg);
  cvReleaseImage(&velx);
  cvReleaseImage(&vely);
}

//---------------------------------------------------------------------------
//  Clculo de la imagen acumulada
//---------------------------------------------------------------------------

IplImage *acum= NULL;    // Imagen acumulada
double actx, acty;       // Posicin actual dentro de la acumulada

//-------------------------

void inic_acum (int xsize, int ysize, int posx, int posy, IplImage *prime)
// Inicializa la imagen acum, dado el primer frame. El tamao de acum es
// (actx X acty), y la posicin inicial es (posx, posy)
{
  CvRect roi;
  acum= cvCreateImage(cvSize(xsize, ysize), IPL_DEPTH_8U, 3);
  actx= posx;
  acty= posy;
  cvSet(acum, cvScalarAll(255));
  roi.x= prime->width/2-20;
  roi.y= 0;
  roi.width= prime->width-1-roi.x;
  roi.height= prime->height;
  cvSetImageROI(prime, roi);
  roi.x= actx+0.5;
  roi.y= acty+0.5;
  cvSetImageROI(acum, roi);
  cvCopy(prime, acum);
  cvResetImageROI(acum);
  cvResetImageROI(prime);
  cvNamedWindow("Acumulada", 0);
  cvShowImage("Acumulada", acum);
  cvResizeWindow("Acumulada", 1000, 260);
}

//-------------------------

void acumula (IplImage *nueva, double dx, double dy)
// Acumula una nueva imagen, que tiene desplazamiento (dx, dy) respecto
// a la anterior
{
  CvRect roi;
  actx+= dx;
  acty+= dy;
  roi.x= nueva->width/2-20;
  roi.y= 0;
  roi.width= izquierda?abs(dx)+1:BLOQUE;
  roi.height= nueva->height;
  cvSetImageROI(nueva, roi);
  roi.x= actx;
  roi.y= acty;
  cvSetImageROI(acum, roi);
  cvCopy(nueva, acum);
  cvResetImageROI(nueva);
  cvResetImageROI(acum);
  cvNamedWindow("Acumulada", 0);
  cvShowImage("Acumulada", acum);
}

//---------------------------------------------------------------------------
//  Procedimiento principal (evento del ratn)
//---------------------------------------------------------------------------

void __fastcall TForm1::Button1Click(TObject *Sender)
{
  int tecla= -1;
  double dx, dy;
  IplImage *ant, *act, *cop;    // Frame anterior, actual y copia del actual
  IplImage *mant, *mact, *mcop; // Lo mismo, pero reducidos
  izquierda= RadioGroup1->ItemIndex==0;

  // 1. Abrir el archivo de vdeo
  CvCapture* acap= cvCaptureFromAVI(Edit1->Text.c_str());
  if (!acap) {
    ShowMessage("El archivo no se encuentra o no es un AVI vlido.");
    return;
  }
  act= cvQueryFrame(acap);
  if (!act) {
    ShowMessage("El archivo no se encuentra o no es un AVI vlido.");
    cvReleaseCapture(&acap);
    return;
  }
  VUELTA(act);

  // 2. Crear las imgenes auxiliares
  mact= cvCreateImage(cvSize(act->width*CSpinEdit1->Value/100,
                     act->height*CSpinEdit1->Value/100), IPL_DEPTH_8U, 3);
  cvResize(act, mact, CV_INTER_LINEAR);
  mcop= cvCloneImage(mact);
  mant= cvCloneImage(mact);
  cop= cvCloneImage(act);
  ant= cvCloneImage(act);

  // 3. Inicializar la imagen acumulada
  act= cvQueryFrame(acap);
  if (!act) {
    ShowMessage("El archivo AVI debe contener como mnimo 2 imgenes.");
    cvReleaseCapture(&acap);
    return;
  }
  VUELTA(act);
  cvResize(act, mact, CV_INTER_LINEAR);
  inic_acum(mact->width*12, mact->height*2,
            izquierda ? mact->width*11:mact->width*1, mact->height*0.5, mact);

  // 4. Para cada frame del vdeo
  do {
    Application->ProcessMessages();
    // 4.1. Calcular el flujo ptico con el frame anterior
    cvCopy(act, cop);
    cvCopy(mact, mcop);
    flujoBM(mact, mant, dx, dy);
    // 4.2. Acumular el frame segn el flujo resultante
    acumula(mact, dx, dy);
    // 4.3. Pasar al siguiente frame
    cvCopy(act, ant);
    cvCopy(mact, mant);
    act= cvQueryFrame(acap);
    if (act) {
      VUELTA(act);
      cvResize(act, mact, CV_INTER_LINEAR);
    }
    tecla= tolower(cvWaitKey(1));
  } while (act && tecla!='q' && tecla!=27);

  // 5. Guardar el resultado y liberar todas las imgenes
  cvSaveImage(Edit2->Text.c_str(), acum);
  cvReleaseCapture(&acap);
  cvReleaseImage(&cop);
  cvReleaseImage(&ant);
  cvReleaseImage(&mcop);
  cvReleaseImage(&mant);
  cvReleaseImage(&mact);
  cvReleaseImage(&acum);
  cvDestroyWindow("Flujo ptico");
}
//---------------------------------------------------------------------------

